import filecmp
import os
import shutil
import tempfile
from inspect import currentframe, getframeinfo
from pathlib import Path

# return the itens in array that is not inexisting or empty
def validate_files(paths):
    valid_files = [path for path in paths if file_is_valid(path)]
    return valid_files

# check if a file is existing and not empty
def file_is_valid(path):
    if path is not None and path.is_file():
        size = path.stat().st_size
        if size > 0:
            return True
    return False

# validate if an string is a valid dir or path with valid content
def check_valid_path(path_str):
    # create a Path object from the input string
    path = Path(path_str)

    # check if the path exists
    if not path.exists():
        raise FileNotFoundError(f"The path '{path_str}' does not exist.")

    # check if it's a directory with at least one file or a valid file with content
    if path.is_dir():
        # check if the directory has at least one file with content
        files_with_content = [file for file in path.iterdir() if file.is_file() and file.stat().st_size > 0]
        if not files_with_content:
            raise ValueError(f"The directory '{path_str}' does not contain any files with content.")
    elif path.is_file():
        # check if the file has content
        if path.stat().st_size == 0:
            raise ValueError(f"The file '{path_str}' does not contain any content.")
    else:
        raise ValueError(f"The path '{path_str}' is neither a valid directory nor a valid file.")

    return path

# validate if an string is a valid dir or path with valid content
def check_existing_path(path_str):
    # create a Path object from the input string
    path = Path(path_str)

    # check if the path exists
    if not path.exists():
        raise FileNotFoundError(f"The path '{path_str}' does not exist.")

    # check if it's a directory with at least one file or a valid file with content
    if not path.is_dir() and not path.is_file():
        raise ValueError(f"The path '{path_str}' is neither a valid directory nor a valid file.")

    return path_str

# create a tempfile class to use as object
class TempFile:

    def __init__(self, final_path: Path, file_ext: str = None):
        self.final_path: Path = None if final_path is None else Path(
            final_path)
        self.file_ext = file_ext
        os.makedirs(Path(Path(getframeinfo(
            currentframe()).filename).resolve().parent, "temp"), exist_ok=True)
        self.temp_file: tempfile.NamedTemporaryFile = tempfile.NamedTemporaryFile(dir=Path(Path(getframeinfo(currentframe()).filename).resolve().parent, "temp"),
                                                                                  delete=False, suffix=file_ext)

        self.temp_file_name = self.temp_file.name
        self.temp_file_path: Path = Path(self.temp_file.name)

        self.temp_file.close()

    # return the actual path of the file
    def getpath(self):
        if self.temp_file_path.is_file():
            return self.temp_file_path
        elif file_is_valid(self.final_path):
            return self.final_path
        else:
            return None

    # return the actual path of the file if not empty
    def getvalidpath(self):
        if file_is_valid(self.temp_file_path):
            return self.temp_file_path
        elif file_is_valid(self.final_path):
            return self.final_path
        else:
            return None

    # save the temp file into final path
    def save(self, overwrite_if_valid: bool = True, update_path: Path = None):
        # update final path case user specifies it
        if update_path is None:
            path: Path = self.final_path
        else:
            path: Path = update_path

        try:
            # if file not valid ou overwrite is enabled, move overwiting existing file
            if not file_is_valid(self.final_path) or overwrite_if_valid:
                os.makedirs(path.parent, exist_ok=True)
                shutil.move(self.temp_file_path, path)
                self.final_path = path
        except Exception as e:
            print(f"Error saving file: {e}")
            return

    # delete the file
    def destroy(self):
        try:
            # destroy temporary file if it exists
            if file_is_valid(self.temp_file_path):
                os.remove(self.temp_file_path)
        except Exception:
            return

# copy an source file to destination if destination file is not equals to source
def copy_file_if_different(src_file: Path, dst_file: Path, silent: bool = False):
    if file_is_valid(dst_file):
        # Check if destination file exists and is different from source file
        if filecmp.cmp(src_file, dst_file) and not silent:
            print(f"{dst_file} already exists and is the same. No need to copy.")
            return

    os.makedirs(dst_file.parent, exist_ok=True)
    shutil.copyfile(src_file, dst_file)
    if not silent:
        print(f"copied to {dst_file}")

# function to delete dir and all its content using shutil
def delete_folder(path: Path):
    if path.is_dir():
        shutil.rmtree(path)


def update_folder_times(folder_path):
    folder_path = Path(folder_path)

    # Keep track of the newest file's modification time for this folder
    newest_file_time = None

    for item in folder_path.iterdir():
        # Check if the item is a file
        if item.is_file():
            file_time = item.stat().st_mtime  # Get modification time of the file

            # Update the newest_file_time if it's the first file or if the current file is newer
            if newest_file_time is None or file_time > newest_file_time:
                newest_file_time = file_time

        # If the item is a subfolder, recursively update its times and find its newest file time
        elif item.is_dir():
            subfolder_newest_time = update_folder_times(item)

            # Update the newest_file_time if a subfolder's newest file is newer
            if newest_file_time is None or (subfolder_newest_time is not None and subfolder_newest_time > newest_file_time):
                newest_file_time = subfolder_newest_time

    # Update the folder's modification and creation times with the newest_file_time
    if newest_file_time is not None:
        # Convert to an integer (necessary on some systems)
        newest_file_time = int(newest_file_time)
        os.utime(path=folder_path, times=(newest_file_time, newest_file_time))

    return newest_file_time  # Return the newest_file_time to update the parent folder's time



import asyncio
import os
from pathlib import Path

import deep_translator
import pysrt
import tqdm.asyncio
import subtitle_utils
from utils import format_time
from gemini_utils import (
    GeminiTranslationConfig,
    normalize_api_keys,
    translate_with_gemini,
)

# all entence endings for japanese and normal people languages
sentence_endings = ['.', '!', '?', ')', 'よ', 'ね',
                    'の', 'さ', 'ぞ', 'な', 'か', '！', '。', '」', '…']

# a good separator is a char or string that doenst change the translation quality but is near ever preserved in result at same or near position
separator = " ◌ "
separator_unjoin = separator.replace(' ', '')
chunk_max_chars = 4999


def translate_srt_file(
    srt_file_path: Path,
    translated_subtitle_path: Path,
    target_lang,
    translate_engine: str = "google",
    gemini_api_keys=None
):
    """
    Translate SRT file using the specified engine.
    translate_engine: "google" or "gemini"
    gemini_api_keys: optional sequence of API keys required if translate_engine == "gemini"
    """
    # Load the original SRT file
    subs = pysrt.open(srt_file_path, encoding='utf-8')

    # Extract the subtitle content and store it in a list. Also rejoin all lines splited
    sub_content = [' '.join(sub.text.strip().splitlines()) for sub in subs]

    if translate_engine == "gemini":
        api_keys = normalize_api_keys(gemini_api_keys)
        if not api_keys:
            raise ValueError("Gemini API key is required for Gemini translation. Get one at https://aistudio.google.com/apikey")

        Path(translated_subtitle_path).unlink(missing_ok=True)

        subs = translate_with_gemini(
            GeminiTranslationConfig(
                api_keys=api_keys,
                input_file=srt_file_path,
                output_file=translated_subtitle_path,
                target_language=target_lang,
            )
        )

        return subs

    # Default: Google Translate
    # Make chunks of at maximum $chunk_max_chars to stay under Google Translate public API limits
    chunks = join_sentences(sub_content, chunk_max_chars) or []

    # Empty list to store enumerated translated chunks
    translated_chunks = [None] * len(chunks)

    tasks = []
    # Limit to 7 concomitant running tasks
    semaphore = asyncio.Semaphore(7)

    # Async chunks translate function
    async def translate_async():
        async def run_translate(index, chunk, lang):
            while True:
                try:
                    async with semaphore:
                        result = await asyncio.wait_for(translate_chunk(index, chunk, lang), 120)
                    translated_chunks[index] = result
                    break
                except Exception:
                    # Restart task
                    await asyncio.sleep(3)

        for index, chunk in enumerate(chunks):
            task = asyncio.create_task(
                run_translate(index, chunk, target_lang))
            tasks.append(task)

        for tsk in tqdm.asyncio.tqdm_asyncio.as_completed(tasks, total=len(tasks), desc="Translating", unit="chunks", unit_scale=False, leave=True, bar_format="{desc} {percentage:3.0f}% | {n_fmt}/{total_fmt} | ETA: {remaining} | ⏱: {elapsed}"):
            await tsk

    # Cria um loop de eventos e executa as tasks
    loop = asyncio.get_event_loop()
    loop.run_until_complete(translate_async())

    print('Processing translation...', end='')

    # Unjoin lines within each chunk that end with a sentence ending
    unjoined_texts = [unjoin_sentences(
        chunk, translated_chunks[i], separator_unjoin) or "" for i, chunk in enumerate(chunks)]
    unjoined_texts = [text for sublist in unjoined_texts for text in sublist]

    # Split lines as necessary targeting same number of lines as original string
    for i, segment in enumerate(unjoined_texts):
        unjoined_texts[i] = "\n".join(subtitle_utils.split_string_to_max_lines(
            text=segment, max_width=0, max_lines=len(subs[i].text.splitlines())))

    # Combine the original and translated subtitle content
    for i, sub in enumerate(subs):
        sub.text = unjoined_texts[i]

    # Save the translated SRT file
    os.makedirs(translated_subtitle_path.parent, exist_ok=True)
    subs.save(translated_subtitle_path, encoding='utf-8')

    print('\r                         ', end='\r')

    return subs

# Async chunk translate function


async def translate_chunk(index, chunk, target_lang):
    while True:
        try:
            # Translate the subtitle content of the chunk using Google Translate
            translator = deep_translator.google.GoogleTranslator(
                source='auto', target=target_lang)
            translated_chunk: str = await asyncio.wait_for(asyncio.get_event_loop().run_in_executor(None, translator.translate, chunk), 30)
            await asyncio.sleep(0)

            # if nothing is retuned, return the original chunk
            if translated_chunk is None or len(translated_chunk.replace(separator.strip(), '').split()) == 0:
                return chunk

            return translated_chunk
        except Exception as e:
            # If an error occurred, retry
            del translator
            print(
                f"\r[chunk {index}]: Exception: {e.__doc__} Retrying in 30 seconds...", flush=True)
            await asyncio.sleep(30)


def join_sentences(lines, max_chars):
    """Join sentences in chunks that stay under *max_chars* without breaking the separator mapping."""
    joined_lines = []
    current_chunk = ""

    for index, line in enumerate(lines):
        if not line:
            line = '\u3164'  # invisible char (not a simple space)

        addition = line + separator

        # if adding the current line would overflow, flush the chunk first
        if current_chunk and len(current_chunk) + len(line) + len(separator) > max_chars:
            joined_lines.append(current_chunk)
            current_chunk = ""

        if len(addition) > max_chars:
            # a single line exceeds the limit; truncate conservatively
            end_index = line.rfind(' ', 0, max_chars - (1 + len(separator)))
            if end_index == -(1 + len(separator)):
                end_index = max_chars - (1 + len(separator))
            joined_lines.append((line[:end_index] + '\u2026' + separator)[:max_chars])
            continue

        current_chunk += addition

        is_last_line = index == len(lines) - 1
        ends_sentence = any(line.endswith(ending) for ending in sentence_endings)

        if not ends_sentence and not is_last_line:
            continue

        if is_last_line:
            joined_lines.append(current_chunk)
            current_chunk = ""
            continue

        next_line = lines[index + 1] or '\u3164'
        next_addition_length = len(next_line) + len(separator)

        if len(current_chunk) + next_addition_length > max_chars:
            joined_lines.append(current_chunk)
            current_chunk = ""

    if current_chunk:
        joined_lines.append(current_chunk)

    return joined_lines


def unjoin_sentences(original_sentence: str, modified_sentence: str, separator: str):
    """
    Splits the original and modified sentences into lines based on the separator.
    Tries to match the number of lines between the original and modified sentences.
    """

    if original_sentence is None:
        return ' '

    # split by separator, remove double spaces and empty or only space strings from list
    original_lines = original_sentence.split(separator)
    original_lines = [s.strip().replace('  ', ' ').lstrip(" ,.:;)") if s.strip().replace('  ', ' ').lstrip(" ,.:;)") else s
                      for s in original_lines if s.strip()]
    original_lines = [s for s in original_lines if s]
    original_lines = [s for s in original_lines if s.strip()]

    if modified_sentence is None:
        return original_lines or ' '

    # fix strange formatation returned by google translate, case occuring
    modified_sentence = modified_sentence.replace(f"{separator_unjoin} ", f"{separator_unjoin}").replace(
        f" {separator_unjoin}", f"{separator_unjoin}").replace(
        f"{separator_unjoin}.", f".{separator_unjoin}").replace(f"{separator_unjoin},", f",{separator_unjoin}")

    # split by separator, remove double spaces and empty or only space strings from list
    modified_lines = modified_sentence.split(separator_unjoin)
    modified_lines = [s.strip().replace('  ', ' ').lstrip(" ,.:;)") if s.strip().replace('  ', ' ').lstrip(" ,.:;)") else s
                      for s in modified_lines if s.strip()]
    modified_lines = [s for s in modified_lines if s]
    modified_lines = [s for s in modified_lines if s.strip()]

    # if original lines is "silence" sign, doenst translate
    if original_lines == "..." or original_lines == "…":
        return original_lines

    # all ok, return lines
    if len(original_lines) == len(modified_lines):
        return modified_lines

    # zero words? return original sentence, removing separator
    original_word_count = sum(len(line.strip().split())
                              for line in original_lines)
    modified_word_count = len(' '.join(modified_lines).strip().split())
    if original_word_count == 0 or modified_word_count == 0:
        return original_sentence.replace(separator, ' ').replace('  ', ' ')

    # calculate proportion of words between original and translated
    modified_words_proportion = modified_word_count / original_word_count
    # list all modified words
    modified_words = ' '.join(modified_lines).replace(separator, "").replace(
        separator_unjoin, "").replace("  ", " ").strip().split(' ')

    new_modified_lines = []
    current_index = 0

    # reconstruct lines based on proportion of original and translated words
    for i in range(len(original_lines)):
        # Calculate the number of words for the current modified sentence
        num_words = int(
            round(len(original_lines[i].strip().split()) * modified_words_proportion))

        # Extract words from modified list
        generated_line = ' '.join(
            modified_words[current_index:current_index+num_words])

        # Update the current index
        current_index += num_words

        # append remaining if is the last loop
        if i == len(original_lines) - 1:
            tail = ' '.join(modified_words[current_index:])
            if tail:
                generated_line = ' '.join([generated_line, tail]).strip()
            current_index = len(modified_words)

        # Add modified sentence to the new list
        new_modified_lines.append(generated_line.replace("  ", " ").strip())

    # case it continues being shorter
    while len(new_modified_lines) < len(original_lines):
        new_modified_lines.append(new_modified_lines[-1])

    return new_modified_lines or original_lines or ' '


import os
import tempfile
import time
import unittest
from pathlib import Path

import file_utils


class FileUtilsTests(unittest.TestCase):
    def test_validate_files_filters_invalid(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            valid = tmp_path / "valid.txt"
            empty = tmp_path / "empty.txt"
            missing = tmp_path / "missing.txt"

            valid.write_text("data", encoding="utf-8")
            empty.write_text("", encoding="utf-8")

            result = file_utils.validate_files([valid, empty, missing])
            self.assertEqual(result, [valid])

    def test_tempfile_lifecycle(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            final_path = tmp_path / "final.txt"

            temp_file = file_utils.TempFile(final_path, file_ext=".txt")
            temp_path = temp_file.getpath()
            self.assertTrue(temp_path.exists())

            temp_path.write_text("temporary", encoding="utf-8")
            temp_file.save()

            self.assertTrue(final_path.exists())
            self.assertEqual(final_path.read_text(encoding="utf-8"), "temporary")

            temp_file.destroy()  # should not raise even if nothing to delete

    def test_copy_file_if_different(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            src = tmp_path / "src.txt"
            dst = tmp_path / "dst.txt"

            src.write_text("content", encoding="utf-8")
            file_utils.copy_file_if_different(src, dst, silent=True)

            self.assertTrue(dst.exists())
            self.assertEqual(dst.read_text(encoding="utf-8"), "content")

    def test_update_folder_times_tracks_newest_file(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            folder = Path(tmpdir)
            older = folder / "older.txt"
            newer = folder / "newer.txt"

            older.write_text("old", encoding="utf-8")
            time.sleep(0.01)
            newer.write_text("new", encoding="utf-8")

            original_mtime = folder.stat().st_mtime
            newest = file_utils.update_folder_times(folder)

            self.assertGreaterEqual(newest, int(newer.stat().st_mtime))
            self.assertGreaterEqual(int(folder.stat().st_mtime), int(original_mtime))

    def test_check_existing_path_accepts_files_and_directories(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            folder = Path(tmpdir)
            file_path = folder / "file.txt"
            file_path.write_text("data", encoding="utf-8")

            self.assertEqual(file_utils.check_existing_path(str(folder)), str(folder))
            self.assertEqual(file_utils.check_existing_path(str(file_path)), str(file_path))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



import importlib
import unittest

try:
    gemini_utils = importlib.import_module("gemini_utils")
except ModuleNotFoundError:  # pragma: no cover - dependency chain might be missing
    gemini_utils = None


@unittest.skipIf(gemini_utils is None, "gemini_utils dependencies are unavailable")
class GeminiUtilsTests(unittest.TestCase):
    def test_normalize_api_keys_handles_iterables_and_duplicates(self):
        keys = [" key1 ", "key2,key3", "key1", None, "\nkey4\n"]
        expected = ["key1", "key2", "key3", "key4"]
        self.assertEqual(gemini_utils.normalize_api_keys(keys), expected)

    def test_normalize_api_keys_accepts_single_string(self):
        self.assertEqual(gemini_utils.normalize_api_keys("key"), ["key"])

    def test_normalize_api_keys_empty_input(self):
        self.assertEqual(gemini_utils.normalize_api_keys(None), [])
        self.assertEqual(gemini_utils.normalize_api_keys([]), [])


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Sequence

import pysrt

import gemini_srt_translator as gst


@dataclass(frozen=True)
class GeminiTranslationConfig:
    api_keys: Sequence[str]
    input_file: Path
    output_file: Path
    target_language: str
    batch_size: int = 400
    temperature: float = 0.3
    top_p: float = 0.9
    top_k: int = 50
    free_quota: bool = True
    resume: bool = False
    thinking: bool = False
    progress_log: bool = False
    thoughts_log: bool = False


class MultiKeyGeminiTranslator(gst.GeminiSRTTranslator):
    """Gemini translator that rotates across an arbitrary number of API keys."""

    def __init__(self, api_keys: Sequence[str], **kwargs) -> None:
        cleaned: List[str] = [key.strip() for key in api_keys if key and key.strip()]
        if not cleaned:
            raise ValueError("At least one Gemini API key is required.")

        primary = cleaned[0]
        secondary = cleaned[1] if len(cleaned) > 1 else None

        super().__init__(gemini_api_key=primary, gemini_api_key2=secondary, **kwargs)

        self._api_keys = cleaned
        self._api_index = 0
        self.current_api_key = primary
        self.current_api_number = 1
        self.backup_api_number = 2 if len(cleaned) > 1 else 1

    def _switch_api(self) -> bool:  # type: ignore[override]
        if len(self._api_keys) <= 1:
            return False

        previous_number = self.current_api_number
        total_keys = len(self._api_keys)

        for step in range(1, total_keys + 1):
            next_index = (self._api_index + step) % total_keys
            if next_index == self._api_index:
                continue

            next_key = self._api_keys[next_index]
            if next_key:
                self._api_index = next_index
                self.current_api_key = next_key
                self.current_api_number = next_index + 1
                self.backup_api_number = previous_number
                return True

        return False


def translate_with_gemini(config: GeminiTranslationConfig) -> pysrt.SubRipFile:
    translator = MultiKeyGeminiTranslator(
        api_keys=config.api_keys,
        target_language=config.target_language,
        input_file=str(config.input_file),
        output_file=str(config.output_file),
        batch_size=config.batch_size,
        temperature=config.temperature,
        top_p=config.top_p,
        top_k=config.top_k,
        free_quota=config.free_quota,
        resume=config.resume,
        thinking=config.thinking,
        progress_log=config.progress_log,
        thoughts_log=config.thoughts_log,
    )

    translator.translate()

    return pysrt.open(config.output_file, encoding="utf-8")


def normalize_api_keys(keys: Iterable[str] | str | None) -> List[str]:
    if keys is None:
        return []

    if isinstance(keys, str):
        raw = [keys]
    else:
        raw = list(keys)

    candidates: List[str] = []
    for value in raw:
        if not value:
            continue
        parts = [part.strip() for part in str(value).replace("\n", ",").split(",")]
        candidates.extend(part for part in parts if part)

    unique: List[str] = []
    seen = set()
    for key in candidates:
        if key not in seen:
            unique.append(key)
            seen.add(key)

    return unique



import time
from pathlib import Path

from contextlib import contextmanager


@contextmanager
def time_task(message_start=None, end=' ', message="⏱ Took"):
    if message_start:
        print(message_start, end=end, flush=True)
    start_time = time.time()
    yield
    end_time = time.time()
    elapsed_time = end_time - start_time
    formatted_elapsed_time = format_time(elapsed_time)
    print(f"{message} {formatted_elapsed_time}", flush=True)


def format_time(elapsed_time):
    hours, rem = divmod(elapsed_time, 3600)
    minutes, seconds = divmod(rem, 60)

    # Check and format the non-zero time units
    parts = []
    if hours:
        parts.append(f"{int(hours)}h")
    if minutes:
        parts.append(f"{int(minutes)}m")
    if seconds or not parts:
        # Add seconds if it's the only non-zero unit or if all units are zero
        parts.append(f"{int(seconds)}s")

    return ' '.join(parts)


def time_func(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        formatted_elapsed_time = format_time(elapsed_time)
        print(f"Execution time of {func.__name__}: {formatted_elapsed_time}")
        return result
    return wrapper

def check_other_extensions(file_path, extensions_to_check):
    """
    Check the existence of files with the same name but different extensions
    in the same folder.

    Parameters:
    - file_path (str): The path of the file to check.
    - extensions_to_check (list): List of extensions to check.

    Returns:
    - list: List of existing file paths with different extensions.
    """
    file_path = Path(file_path)
    folder = file_path.parent
    base_name = file_path.stem

    matching_files = [
        folder / (base_name + ext)
        for ext in extensions_to_check
        if (folder / (base_name + ext)).exists()
    ]

    return matching_files

video_extensions = {".mp4", ".webm", ".mkv", ".avi", ".mov", ".wmv", ".flv", ".vob", ".mts", ".m2ts", ".ts", ".yuv", ".mpg", ".mp2", ".mpeg", ".mpe", ".mpv", ".m2v", ".m4v", ".3gp", ".3g2", ".nsv", ".mts"}
audio_extensions = {".aa", ".aac", ".aax", ".act", ".aiff", ".alac", ".amr", ".ape", ".au", ".awb", ".dss", ".dvf", ".flac", ".gsm", ".iklax", ".ivs", ".m4a", ".m4b", ".m4p", ".mpga", ".mmf", ".mp3", ".mpc", ".msv", ".nmf", ".ogg", ".oga", ".mogg", ".opus", ".ra", ".rm", ".raw", ".rf64", ".sln", ".tta", ".voc", ".vox", ".wav", ".wma", ".wv", ".webm", ".8svx"}


import subprocess
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

import ffmpeg_utils


class DummyTqdm:
    def __init__(self, *_, **__):
        self.n = 0

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return False

    def update(self, value):
        self.n += value


class DummyProgress:
    def __init__(self, cmd):
        self.cmd = cmd

    def run_command_with_progress(self):
        yield 100


class FfmpegUtilsTests(unittest.TestCase):
    def test_existing_subtitles_mapped_after_generated(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            input_media = Path(tmpdir) / "input.mp4"
            input_media.write_bytes(b"media")

            srt_file = Path(tmpdir) / "legen.srt"
            srt_file.write_text("1\n00:00:00,000 --> 00:00:01,000\ntext\n", encoding="utf-8")

            output_media = Path(tmpdir) / "output.mp4"
            captured_cmd = {}

            def fake_run(command, capture_output=False, text=False, **kwargs):
                if "-show_streams" in command:
                    return subprocess.CompletedProcess(command, 0, stdout="DISPOSITION:attached_pic=0")
                raise AssertionError(f"Unexpected command: {command}")

            with patch("ffmpeg_utils.subprocess.run", side_effect=fake_run), \
                 patch("ffmpeg_utils.tqdm", side_effect=lambda *a, **k: DummyTqdm()), \
                 patch("ffmpeg_utils.FfmpegProgress", side_effect=lambda cmd: DummyProgress(cmd)) as mock_progress:
                ffmpeg_utils.insert_subtitle(
                    input_media_path=input_media,
                    subtitles_path=[srt_file],
                    burn_subtitles=False,
                    output_video_path=output_media,
                    codec_video="h264",
                    codec_audio="aac",
                )

            cmd = mock_progress.call_args[0][0]
            map_entries = [cmd[i + 1] for i, arg in enumerate(cmd) if arg == "-map"]

            self.assertIn("0:s?", map_entries)
            generated_indexes = [idx for idx, value in enumerate(map_entries) if value.endswith(":s") and value != "0:s?"]
            self.assertTrue(generated_indexes)
            downloaded_index = map_entries.index("0:s?")
            self.assertTrue(all(idx < downloaded_index for idx in generated_indexes))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



import os
from pathlib import Path

import whisperx_legen_fork as whisperx
import whisper
from whisperx_legen_fork import asr, audio as wx_audio, alignment, utils
import whisper_utils
import subtitle_utils
from utils import time_task

def transcribe_audio(model: asr.WhisperModel, audio_path: Path, srt_path: Path, lang: str = None, device: str = "cpu", batch_size: int = 4):
    audio = wx_audio.load_audio(file=audio_path.as_posix(), sr=model.model.feature_extractor.sampling_rate)
        
    # Define the progress callback function
    def progress_callback(state, current: int = None, total: int = None):
        args = state, current, total
        args = [arg for arg in args if arg is not None]

        if len(args) == 1:
            state = args[0]
        if len(args) > 1:
            total = args[-1]
            current = args[-2]
            state = None
        if len(args) > 2:
            state = args[-3]

        try:
            if state is None:
                state = "WhisperX"  
            elif type(state) == 'String' or type(state) == int:
                state = state
            else:
                state = state.value
        except:
            state = "WhisperX"

        print('\r                                                            \r' + state + ((': ' + str(round(current/total*100)) + '%') if current and total else '') + ((' [' + str(current) + '/' + str(total) + ']') if current and total else ''), end=' ', flush=True)

    # Transcribe
    with time_task("Running WhisperX transcription engine...", end='\n'):
        transcribe = model.transcribe(audio=audio, language=lang, batch_size=batch_size, on_progress=progress_callback)

    # Align if possible
    if lang in alignment.DEFAULT_ALIGN_MODELS_HF or lang in alignment.DEFAULT_ALIGN_MODELS_TORCH:
        with time_task(message_start="Running alignment...", end='\n'):
            try:
                model_a, metadata = alignment.load_align_model(language_code=lang, device="cuda")
                transcribe = alignment.align(transcript=transcribe["segments"], model=model_a, align_model_metadata=metadata, audio=audio, device="cuda", return_char_alignments=True, on_progress=progress_callback)
            except Exception:
                model_a, metadata = alignment.load_align_model(language_code=lang, device="cpu")  # force load on cpu due errors on gpu
                transcribe = alignment.align(transcript=transcribe["segments"], model=model_a, align_model_metadata=metadata, audio=audio, device="cpu", return_char_alignments=True, on_progress=progress_callback)
    else:
        print(f"Language {lang} not suported for alignment. Skipping this step")

    # Format subtitles
    segments = subtitle_utils.format_segments(transcribe['segments'])

    # Save the subtitle file
    subtitle_utils.SaveSegmentsToSrt(segments, srt_path)

    return transcribe


def detect_language(model: asr.WhisperModel, audio_path: Path):
    try:
        if os.getenv("COLAB_RELEASE_TAG"):
            raise Exception("Method invalid for Google Colab") 
        audio = wx_audio.load_audio(audio_path.as_posix(), model.model.feature_extractor.sampling_rate)
        audio = wx_audio.pad_or_trim(audio, model.model.feature_extractor.n_samples)
        mel = wx_audio.log_mel_spectrogram(audio, n_mels=model.model.model.n_mels)
        encoder_output = model.model.encode(mel)
        results = model.model.model.detect_language(encoder_output)
        language_token, language_probability = results[0][0]
        return language_token[2:-2]
    except:
        print("using whisper base model for detection: ", end='')
        whisper_model = whisper.load_model("base", device="cpu", in_memory=True)
        return whisper_utils.detect_language(model=whisper_model, audio_path=audio_path)



import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Iterable, List


def _resolve_downloader() -> str:
    """Return the executable name for yt-dlp."""
    downloader = "yt-dlp"
    if shutil.which(downloader):
        return downloader
    raise FileNotFoundError(
        "yt-dlp executable not found. Install yt-dlp before using URL downloads."
    )


def _append_downloaded_suffix_to_subtitles(media_path: Path) -> None:
    """Ensure existing subtitle streams are labeled with a [downloaded] suffix."""
    ffprobe_cmd = [
        "ffprobe",
        "-loglevel",
        "error",
        "-select_streams",
        "s",
        "-show_entries",
        "stream=index:stream_tags=language,title",
        "-of",
        "json",
        "file:" + media_path.expanduser().resolve().as_posix(),
    ]

    try:
        probe = subprocess.run(
            ffprobe_cmd,
            check=True,
            capture_output=True,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        raise RuntimeError(f"ffprobe failed while inspecting subtitles: {exc}") from exc

    try:
        data = json.loads(probe.stdout or "{}")
    except json.JSONDecodeError as exc:
        raise RuntimeError("ffprobe returned invalid JSON for subtitle metadata") from exc

    streams = data.get("streams") or []
    if not streams:
        return

    metadata_args: List[str] = []
    update_needed = False

    for subtitle_index, stream in enumerate(streams):
        tags = stream.get("tags") or {}
        title = (tags.get("title") or "").strip()
        language = (tags.get("language") or "").strip()

        base_title = title or language or f"Subtitle {subtitle_index + 1}"
        if not base_title:
            base_title = f"Subtitle {subtitle_index + 1}"

        final_title = (
            base_title
            if base_title.endswith(" [downloaded]")
            else f"{base_title} [downloaded]"
        )

        if title != final_title:
            update_needed = True

        metadata_args.extend([
            f"-metadata:s:s:{subtitle_index}",
            f"title={final_title}",
        ])

    if not update_needed:
        return

    with tempfile.NamedTemporaryFile(
        dir=media_path.parent, suffix=media_path.suffix, delete=False
    ) as tmp:
        temp_output = Path(tmp.name)

    ffmpeg_cmd = [
        "ffmpeg",
        "-y",
        "-i",
        "file:" + media_path.as_posix(),
        "-map",
        "0",
        "-c",
        "copy",
    ]

    ffmpeg_cmd.extend(metadata_args)
    ffmpeg_cmd.extend(["-movflags", "+faststart", "file:" + temp_output.as_posix()])

    try:
        subprocess.run(ffmpeg_cmd, check=True)
    except subprocess.CalledProcessError as exc:
        temp_output.unlink(missing_ok=True)
        raise RuntimeError(f"ffmpeg failed while tagging subtitles: {exc}") from exc

    os.replace(temp_output, media_path)


_FORMAT_SUFFIX_RE = re.compile(r"\.f\d+(?:[a-z0-9_-]+)?$", re.IGNORECASE)


def _title_from_destination(destination: str) -> str:
    """Derive a human-friendly title from a yt-dlp destination path."""
    sanitized = destination.strip().strip('"').strip("'")
    name = Path(sanitized).name
    base, _ = os.path.splitext(name)
    cleaned = _FORMAT_SUFFIX_RE.sub("", base).strip()
    return cleaned or name


def _subtitle_label_from_path(path_str: str) -> str:
    """Extract a readable subtitle label (language or filename)."""
    sanitized = path_str.strip().strip('"')
    if sanitized.startswith("file:"):
        sanitized = sanitized[5:]
    candidate = Path(sanitized)
    suffixes = candidate.suffixes
    if len(suffixes) >= 2:
        lang_suffix = suffixes[-2]
        if lang_suffix.startswith("."):
            lang = lang_suffix[1:]
        else:
            lang = lang_suffix
        lang = lang.strip()
        if lang:
            return lang
    return candidate.name


def download_urls(
    urls: Iterable[str],
    output_dir: Path,
    overwrite: bool = False,
    download_remote_subs: bool = False,
) -> List[Path]:
    """Download one or more media URLs into output_dir using yt-dlp."""
    downloader = _resolve_downloader()
    output_dir = Path(output_dir).expanduser().resolve()
    output_dir.mkdir(parents=True, exist_ok=True)

    format_selector = "bestvideo[ext=mp4]+bestaudio[ext=m4a]/bestvideo[ext=mp4]+bestaudio/bestvideo+bestaudio/best"
    command = [
        downloader,
        "--no-warnings",
        "--newline",
        "--progress",
        "--format",
        format_selector,
        "--merge-output-format",
        "mp4",
        "-P",
        str(output_dir),
        "--continue",
    ]
    if download_remote_subs:
        command.extend([
            "--embed-subs",
            "--sub-langs",
            "all",
        ])
    if overwrite:
        command.append("--force-overwrites")
    else:
        command.append("--no-overwrites")

    url_list = [url.strip() for url in urls if url and url.strip()]
    if not url_list:
        raise ValueError("No URLs provided to download")

    command.extend(url_list)

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    current_destination: Path | None = None
    current_status: str | None = None
    current_title: str | None = None
    current_url: str | None = None
    current_title_printed = False
    current_stream_index = 0
    active_stream_is_primary = False
    progress_active = False
    last_progress_len = 0
    error_lines: List[str] = []

    def current_label(fallback: str | None = None) -> str:
        if current_title:
            return current_title
        if current_url:
            return current_url
        if current_destination is not None:
            return current_destination.name
        return fallback or "download"

    def finish_progress() -> None:
        nonlocal progress_active, last_progress_len
        if progress_active:
            sys.stdout.write("\r")
            sys.stdout.write(" " * last_progress_len)
            sys.stdout.write("\r")
            sys.stdout.flush()
            last_progress_len = 0
            progress_active = False

    def update_progress(text: str) -> None:
        nonlocal progress_active, last_progress_len
        try:
            terminal_width = os.get_terminal_size().columns
        except OSError:
            terminal_width = 120

        if terminal_width > 0 and len(text) > terminal_width:
            chunks = [text[i:i + terminal_width] for i in range(0, len(text), terminal_width)]
            finish_progress()
            for chunk in chunks[:-1]:
                print(chunk, flush=True)
            text = chunks[-1]

        display_text = text
        if len(display_text) < last_progress_len:
            display_text += " " * (last_progress_len - len(display_text))

        sys.stdout.write("\r" + display_text)
        sys.stdout.flush()
        last_progress_len = len(display_text)
        progress_active = True

    try:
        assert process.stdout is not None
        for raw_line in process.stdout:
            if raw_line is None:
                continue
            line = raw_line.rstrip("\n")
            stripped = line.strip()
            if not stripped:
                continue

            if "Extracting URL:" in stripped:
                url = stripped.split("Extracting URL:", 1)[1].strip()
                if url:
                    current_url = url
                    current_title = None
                    current_destination = None
                    current_status = None
                    current_title_printed = False
                    current_stream_index = 0
                    active_stream_is_primary = False
                    update_progress(f"Loading [{url}]")
                continue

            destination_prefix = "[download] Destination:"
            resume_prefix = "[download] Resuming download"
            already_downloaded_text = "has already been downloaded"

            if stripped.startswith(destination_prefix):
                destination_value = stripped[len(destination_prefix):].strip()
                target = Path(destination_value)
                if not target.is_absolute():
                    target = (output_dir / target).resolve()
                current_destination = target
                candidate_title = _title_from_destination(destination_value)
                if current_title != candidate_title:
                    current_title = candidate_title
                    current_title_printed = False
                    current_stream_index = 0

                is_primary_stream = current_stream_index == 0

                finish_progress()
                if is_primary_stream:
                    display_name = current_label(target.name)
                    if target.exists() and not overwrite:
                        print(f"Existing: {display_name}", flush=True)
                        current_status = "skip"
                    else:
                        print(f"Downloading: {display_name}", flush=True)
                        current_status = "download"
                    current_title_printed = True
                else:
                    if current_status is None:
                        current_status = "download"

                current_stream_index += 1
                active_stream_is_primary = is_primary_stream
                continue

            if stripped.startswith(resume_prefix):
                if active_stream_is_primary and current_status != "continue":
                    finish_progress()
                    display_name = current_label()
                    print(f"Downloading: {display_name}", flush=True)
                    current_status = "continue"
                continue

            if already_downloaded_text in stripped:
                remainder = stripped.split("[download]", 1)[-1].strip()
                name_part = remainder.split(already_downloaded_text, 1)[0].strip().rstrip('.')
                if not name_part and current_destination is not None:
                    name_part = current_destination.name
                if name_part:
                    title_guess = _title_from_destination(name_part)
                    if title_guess:
                        current_title = title_guess
                finish_progress()
                display_name = current_label(name_part or None)
                if not current_title_printed:
                    print(f"Existing: {display_name}", flush=True)
                    current_title_printed = True
                current_status = "skip"
                active_stream_is_primary = False
                continue

            if download_remote_subs:
                if "Writing video subtitles to:" in stripped or "Writing subtitles to:" in stripped:
                    _, target_part = stripped.split("to:", 1)
                    subtitle_target = target_part.strip()
                    finish_progress()
                    label = _subtitle_label_from_path(subtitle_target)
                    print(f"Downloading subtitle track: {label}", flush=True)
                    continue
                if stripped.lower().startswith("[download] downloading subtitle"):
                    finish_progress()
                    print(stripped, flush=True)
                    continue

            if stripped.startswith("[download]") and "Downloading webpage" in stripped:
                finish_progress()
                current_title = None
                current_title_printed = False
                current_stream_index = 0
                active_stream_is_primary = False
                current_status = None
                continue

            if stripped.startswith("[Merger] Merging formats into"):
                finish_progress()
                current_title = None
                current_title_printed = False
                current_stream_index = 0
                active_stream_is_primary = False
                current_status = None
                continue

            if stripped.startswith("[download]"):
                if "%" in stripped:
                    update_progress(stripped)
                    if "100%" in stripped:
                        finish_progress()
                continue

            if "ERROR" in stripped.upper() or stripped.lower().startswith("error:") or stripped.startswith("[error]"):
                finish_progress()
                print(stripped, flush=True)
                error_lines.append(stripped)

        return_code = process.wait()
        finish_progress()
    except Exception:
        process.kill()
        finish_progress()
        raise

    if return_code != 0:
        if not error_lines:
            print(f"{downloader} reported an error. Review the output above for details.", flush=True)
        raise RuntimeError(
            f"{downloader} failed with exit code {return_code}. Check the terminal output for details."
        )

    downloaded_files = sorted(output_dir.rglob("*.mp4"))
    if not downloaded_files:
        raise RuntimeError(
            f"{downloader} completed without producing MP4 files. Check the URL or provide --overwrite to re-download."
        )

    if download_remote_subs:
        for media_path in downloaded_files:
            try:
                _append_downloaded_suffix_to_subtitles(media_path)
            except RuntimeError as exc:
                print(f"Warning: unable to tag downloaded subtitles in {media_path.name}: {exc}")

    return downloaded_files



import io
import json
import subprocess
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

import download_utils


class DownloadUtilsTests(unittest.TestCase):
    class FakeProcess:
        def __init__(self, output_lines, on_wait=None, returncode=0):
            self.stdout = io.StringIO("".join(output_lines))
            self._returncode = returncode
            self.returncode = returncode
            self._on_wait = on_wait

        def wait(self):
            if self._on_wait:
                self._on_wait()
            return self._returncode

        def kill(self):
            pass

    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_resolve_downloader_found(self, mock_which):
        self.assertEqual(download_utils._resolve_downloader(), "yt-dlp")
        mock_which.assert_called_once_with("yt-dlp")

    @patch("download_utils.shutil.which", return_value=None)
    def test_resolve_downloader_missing(self, mock_which):
        with self.assertRaises(FileNotFoundError):
            download_utils._resolve_downloader()
        mock_which.assert_called_once_with("yt-dlp")

    @patch("download_utils._append_downloaded_suffix_to_subtitles")
    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_resume_and_skip_with_subtitles(self, mock_which, mock_popen, mock_suffix):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir = Path(tmpdir)
            mp4_path = output_dir / "video.mp4"
            (output_dir / "video.mp4.part").write_bytes(b"partial")

            lines = [
                f"[download] Destination: {mp4_path}\n",
                "[download] Resuming download at byte 12345\n",
                "[download]  10.0% of 10.00MiB at 1.00MiB/s ETA 00:10\n",
                "[download] 100% of 10.00MiB in 00:10\n",
            ]

            def on_wait():
                mp4_path.write_bytes(b"data")

            fake_process = self.FakeProcess(lines, on_wait=on_wait)
            captured = {}

            def fake_popen(command, stdout, stderr, text):
                captured["command"] = command
                return fake_process

            mock_popen.side_effect = fake_popen

            result = download_utils.download_urls(
                [" https://example.com/video "],
                output_dir,
                download_remote_subs=True,
            )

            self.assertEqual(result, [mp4_path])
            self.assertTrue(mp4_path.exists())
            self.assertIn("--no-warnings", captured["command"])
            self.assertIn("--progress", captured["command"])
            self.assertIn("--newline", captured["command"])
            self.assertNotIn("--quiet", captured["command"])
            mock_suffix.assert_called_once_with(mp4_path)

    @patch("download_utils._append_downloaded_suffix_to_subtitles")
    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_force_overwrite_with_subtitles(self, mock_which, mock_popen, mock_suffix):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir = Path(tmpdir)
            mp4_path = output_dir / "video.mp4"

            lines = [
                f"[download] Destination: {mp4_path}\n",
                "[download] 100% of 10.00MiB in 00:10\n",
            ]

            def on_wait():
                mp4_path.write_bytes(b"data")

            fake_process = self.FakeProcess(lines, on_wait=on_wait)
            captured = {}

            def fake_popen(command, stdout, stderr, text):
                captured["command"] = command
                return fake_process

            mock_popen.side_effect = fake_popen

            result = download_utils.download_urls(
                ["https://example.com/video"],
                output_dir,
                overwrite=True,
                download_remote_subs=True,
            )

            self.assertEqual(result, [mp4_path])
            self.assertIn("--force-overwrites", captured["command"])
            self.assertNotIn("--no-overwrites", captured["command"])
            mock_suffix.assert_called_once_with(mp4_path)

    @patch("download_utils._append_downloaded_suffix_to_subtitles")
    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_no_subtitles_by_default(self, mock_which, mock_popen, mock_suffix):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir = Path(tmpdir)
            mp4_path = output_dir / "video.mp4"

            lines = [
                f"[download] Destination: {mp4_path}\n",
                "[download] 100% of 10.00MiB in 00:10\n",
            ]

            def on_wait():
                mp4_path.write_bytes(b"data")

            fake_process = self.FakeProcess(lines, on_wait=on_wait)
            captured = {}

            def fake_popen(command, stdout, stderr, text):
                captured["command"] = command
                return fake_process

            mock_popen.side_effect = fake_popen

            result = download_utils.download_urls(["https://example.com/video"], output_dir)

            self.assertEqual(result, [mp4_path])
            self.assertNotIn("--embed-subs", captured["command"])
            self.assertNotIn("--sub-langs", captured["command"])
            mock_suffix.assert_not_called()

    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_no_urls(self, mock_which):
        with self.assertRaises(ValueError):
            download_utils.download_urls([], Path("/tmp"))

    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_multi_stream_prints_title_once(self, mock_which, mock_popen):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir = Path(tmpdir)
            mp4_path = output_dir / "video.mp4"

            lines = [
                "[download] Destination: video.f137.mp4\n",
                "[download] 100% of 10.00MiB in 00:10\n",
                "[download] Destination: video.f140.m4a\n",
                "[download] 100% of 3.00MiB in 00:05\n",
            ]

            def on_wait():
                mp4_path.write_bytes(b"data")

            fake_process = self.FakeProcess(lines, on_wait=on_wait)
            mock_popen.return_value = fake_process

            with patch("sys.stdout", new_callable=io.StringIO) as fake_stdout:
                result = download_utils.download_urls([
                    "https://example.com/video"
                ], output_dir)

            output = fake_stdout.getvalue()
            self.assertEqual(result, [mp4_path])
            self.assertEqual(output.count("Downloading: video"), 1)

    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_subprocess_failure(self, mock_which, mock_popen):
        with tempfile.TemporaryDirectory() as tmpdir:
            fake_process = self.FakeProcess([], returncode=1)
            mock_popen.return_value = fake_process

            with self.assertRaises(RuntimeError) as exc:
                download_utils.download_urls(["https://example.com/video"], Path(tmpdir))
            self.assertIn("failed with exit code", str(exc.exception))

    @patch("download_utils.subprocess.Popen")
    @patch("download_utils.shutil.which", return_value="/usr/bin/yt-dlp")
    def test_download_urls_no_output_files(self, mock_which, mock_popen):
        with tempfile.TemporaryDirectory() as tmpdir:
            fake_process = self.FakeProcess([])
            mock_popen.return_value = fake_process

            with self.assertRaises(RuntimeError) as exc:
                download_utils.download_urls(["https://example.com/video"], Path(tmpdir))
            self.assertIn("without producing MP4 files", str(exc.exception))

    @patch("download_utils.subprocess.run")
    def test_append_downloaded_suffix_no_streams(self, mock_run):
        with tempfile.TemporaryDirectory() as tmpdir:
            media_path = Path(tmpdir) / "video.mp4"
            media_path.write_bytes(b"data")

            mock_run.return_value = subprocess.CompletedProcess(
                ["ffprobe"], 0, stdout=json.dumps({"streams": []})
            )

            download_utils._append_downloaded_suffix_to_subtitles(media_path)
            mock_run.assert_called_once()

    def test_append_downloaded_suffix_updates_metadata(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            media_path = Path(tmpdir) / "video.mp4"
            media_path.write_bytes(b"original")

            streams = {
                "streams": [
                    {"tags": {"title": "English"}},
                    {"tags": {"language": "es"}},
                ]
            }

            def fake_run(command, check=True, capture_output=False, text=False, **kwargs):
                if command[0] == "ffprobe":
                    return subprocess.CompletedProcess(command, 0, stdout=json.dumps(streams))
                if command[0] == "ffmpeg":
                    self.assertIn("-metadata:s:s:0", command)
                    idx0 = command.index("-metadata:s:s:0")
                    self.assertEqual(command[idx0 + 1], "title=English [downloaded]")

                    self.assertIn("-metadata:s:s:1", command)
                    idx1 = command.index("-metadata:s:s:1")
                    self.assertEqual(command[idx1 + 1], "title=es [downloaded]")

                    output = command[-1]
                    if output.startswith("file:"):
                        Path(output[5:]).write_bytes(b"muxed")
                    else:
                        Path(output).write_bytes(b"muxed")
                    return subprocess.CompletedProcess(command, 0)
                raise AssertionError(f"Unexpected command: {command}")

            with patch("download_utils.subprocess.run", side_effect=fake_run) as mock_run:
                download_utils._append_downloaded_suffix_to_subtitles(media_path)

            self.assertEqual(media_path.read_bytes(), b"muxed")
            self.assertEqual(mock_run.call_count, 2)


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



import argparse
import os
import subprocess
import time
from inspect import currentframe, getframeinfo
from pathlib import Path
from urllib.parse import urlparse

import ffmpeg_utils
import file_utils
import download_utils
import subtitle_utils
import translate_utils
from gemini_utils import normalize_api_keys
from utils import time_task, audio_extensions, video_extensions, check_other_extensions

VERSION = "0.19.1"

SUPPORTED_SUBTITLE_FORMATS = {"srt", "txt"}


def looks_like_url(value: str) -> bool:
    if not value:
        return False
    parsed = urlparse(str(value))
    return parsed.scheme in {"http", "https"} and bool(parsed.netloc)


def normalize_subtitle_formats(raw_value):
    if isinstance(raw_value, (list, tuple)):
        values = raw_value
    else:
        values = [raw_value]

    formats = []
    for value in values:
        if value is None:
            continue
        parts = str(value).replace(",", " ").split()
        for part in parts:
            fmt = part.strip().lower()
            if fmt:
                formats.append(fmt)

    if not formats:
        formats = ["srt"]

    # maintain user order while removing duplicates
    seen = {}
    for fmt in formats:
        seen.setdefault(fmt, None)

    return list(seen.keys())

# Terminal colors
default = "\033[1;0m"
gray = "\033[1;37m"
wblue = "\033[1;36m"
blue = "\033[1;34m"
yellow = "\033[1;33m"
green = "\033[1;32m"
red = "\033[1;31m"

print(f"""
{blue}888              {gray} .d8888b.                   
{blue}888              {gray}d88P  Y88b                  
{blue}888              {gray}888    888                  
{blue}888      .d88b.  {gray}888         .d88b.  88888b. 
{blue}888     d8P  Y8b {gray}888  88888 d8P  Y8b 888 "88b
{blue}888     88888888 {gray}888    888 88888888 888  888
{blue}888     Y8b.     {gray}Y88b  d88P Y8b.     888  888
{blue}88888888 "Y8888  {gray} "Y8888P88  "Y8888  888  888

legen {version} - github.com/matheusbach/legen{default}
python {__import__('sys').version}
""")
time.sleep(1.5)

# Define parameters and configurations
parser = argparse.ArgumentParser(prog="LeGen", description="Uses AI to locally transcribes speech from media files, generating subtitle files, translates the generated subtitles, inserts them into the mp4 container, and burns them directly into video",
                                 argument_default=True, allow_abbrev=True, add_help=True, usage='LeGen -i INPUT_PATH [other options]')
parser.add_argument("-i", "--input_path",
                    help="Local media path or URL to download before processing", required=True, type=str)
parser.add_argument("--norm", default=False, action="store_true",
                    help="Normalize folder times and run vidqa on input_path before starting processing files")
parser.add_argument("-ts:e", "--transcription_engine", type=str, default="whisperx",
                    help="Transcription engine. Possible values: whisperx (default), whisper")
parser.add_argument("-ts:m", "--transcription_model", type=str, default="large-v3-turbo",
                    help="Path or name of the Whisper transcription model. A larger model will consume more resources and be slower, but with better transcription quality. Possible values: tiny, base, small, medium, large, turbo, large-v3-turbo (default) ...")
parser.add_argument("-ts:d", "--transcription_device", type=str, default="auto",
                    help="Device to run the transcription through Whisper. Possible values: auto (default), cpu, cuda")
parser.add_argument("-ts:c", "--transcription_compute_type", type=str, default="auto",
                    help="Quantization for the neural network. Possible values: auto (default), int8, int8_float32, int8_float16, int8_bfloat16, int16, float16, bfloat16, float32")
parser.add_argument("-ts:b", "--transcription_batch", type=int, default=4,
                    help="Number of simultaneous segments being transcribed. Higher values will speed up processing. If you have low RAM/VRAM, long duration media files or have buggy subtitles, reduce this value to avoid issues. Only works using transcription_engine whisperx. (default: 4)")
parser.add_argument("--translate", type=str, default="none",
                    help="Translate subtitles to language code if not the same as origin. (default: don't translate)")
parser.add_argument("--translate_engine", type=str, default="google",
                    help="Translation engine to use: google (default) or gemini")
parser.add_argument("--gemini_api_key", action="append", default=[], type=str,
                    help="Gemini API key. Repeat or separate by comma/line break to add multiple keys (required if --translate_engine=gemini)")
parser.add_argument("--input_lang", type=str, default="auto",
                    help="Indicates (forces) the language of the voice in the input media (default: auto)")
parser.add_argument("-c:v", "--codec_video", type=str, default="h264", metavar="VIDEO_CODEC",
                    help="Target video codec. Can be used to set acceleration via GPU or another video API [codec_api], if supported (ffmpeg -encoders). Ex: h264, libx264, h264_vaapi, h264_nvenc, hevc, libx265 hevc_vaapi, hevc_nvenc, hevc_cuvid, hevc_qsv, hevc_amf (default: h264)")
parser.add_argument("-c:a", "--codec_audio", type=str, default="aac", metavar="AUDIO_CODEC",
                    help="Target audio codec. (default: aac). Ex: aac, libopus, mp3, vorbis")
parser.add_argument("-o:s", "--output_softsubs", default=None, type=Path,
                    help="Path to the folder or output file for the video files with embedded softsub (embedded in the mp4 container and .srt files). (default: softsubs_ + input_path)")
parser.add_argument("-o:h", "--output_hardsubs", default=None, type=Path,
                    help="Output folder path for video files with burned-in captions and embedded in the mp4 container. (default: hardsubs_ + input_path)")
parser.add_argument("-o:d", "--output_downloads", default=None, type=Path,
                    help="Destination folder for videos downloaded from URL inputs. (default: ./downloads)")
parser.add_argument("--overwrite", default=False, action="store_true",
                    help="Overwrite existing files in output directories")
parser.add_argument("-dl:rs", "--download_remote_subs", default=False, action="store_true",
                    help="When using a URL input, also download and embed remote subtitle tracks (disabled by default)")
parser.add_argument("--disable_srt", default=False, action="store_true",
                    help="Disable .srt file generation and don't insert subtitles in mp4 container of output_softsubs")
parser.add_argument("--subtitle_formats", type=str, default="srt",
                    help="Subtitle formats to export (separate multiple options with comma or space). Supported: srt, txt")
parser.add_argument("--disable_softsubs", default=False, action="store_true",
                    help="Don't insert subtitles in mp4 container of output_softsubs. This option continues generating .srt files")
parser.add_argument("--disable_hardsubs", default=False, action="store_true",
                    help="Disable subtitle burn in output_hardsubs")
parser.add_argument("--copy_files", default=False, action="store_true",
                    help="Copy other (non-video) files present in input directory to output directories. Only generate the subtitles and videos")
args = parser.parse_args()

input_path_raw = args.input_path

if looks_like_url(input_path_raw):
    download_destination = args.output_downloads or Path.cwd() / "downloads"
    args.output_downloads = Path(download_destination).expanduser().resolve()
    try:
        with time_task(message_start=f"\nDownloading media with yt-dlp to {gray}{args.output_downloads}{default}", end="\n"):
            downloaded_files = download_utils.download_urls(
                [input_path_raw],
                args.output_downloads,
                overwrite=args.overwrite,
                download_remote_subs=args.download_remote_subs,
            )
    except FileNotFoundError as exc:
        parser.error(str(exc))
    except RuntimeError as exc:
        parser.error(str(exc))

    print(f"Downloaded {len(downloaded_files)} file(s) with {gray}yt-dlp{default}. Continuing with local processing from {gray}{args.output_downloads}{default}.")
    args.input_path = args.output_downloads
else:
    candidate_path = Path(input_path_raw).expanduser().resolve()
    if not candidate_path.exists():
        parser.error(f"Input path '{input_path_raw}' is neither an existing file/folder nor a downloadable URL.")
    args.input_path = candidate_path

if args.output_downloads is not None:
    args.output_downloads = Path(args.output_downloads).expanduser().resolve()


requested_formats = normalize_subtitle_formats(args.subtitle_formats)
unsupported_formats = [fmt for fmt in requested_formats if fmt not in SUPPORTED_SUBTITLE_FORMATS]
if unsupported_formats:
    parser.error(f"Unsupported subtitle format(s): {', '.join(unsupported_formats)}. Supported formats: {', '.join(sorted(SUPPORTED_SUBTITLE_FORMATS))}")

if args.disable_srt:
    requested_formats = [fmt for fmt in requested_formats if fmt != "srt"]

if not requested_formats and not args.disable_srt:
    requested_formats = ["srt"]

args.subtitle_formats = requested_formats
args.disable_srt = "srt" not in args.subtitle_formats
args.export_txt = "txt" in args.subtitle_formats


def export_txt_if_requested(source_path: Path, target_path: Path):
    if not args.export_txt:
        return
    if source_path is None:
        return
    if not args.overwrite and file_utils.file_is_valid(target_path):
        return
    subtitle_utils.export_plain_text_from_srt(source_path, target_path)

args.gemini_api_keys = normalize_api_keys(args.gemini_api_key)
args.gemini_api_key = args.gemini_api_keys[0] if args.gemini_api_keys else None

# If Gemini API keys are provided, force translate_engine to 'gemini'
if args.gemini_api_keys:
    args.translate_engine = "gemini"
    # Set pt-BR as default translation language when pt is selected
    if args.translate.lower() == 'pt':
        args.translate = 'pt-BR'
# Fix google compatibility with pt language
if args.translate_engine == "google" and args.translate.lower() in ['pt', 'pt-br', 'pt-pt']:
    args.translate = 'pt'

if not args.output_softsubs and not args.input_path.is_file():
    args.output_softsubs = compatibility_path if (compatibility_path := Path(args.input_path.parent, "legen_srt_" + args.input_path.name)).exists() else Path(args.input_path.parent, "softsubs_" + args.input_path.name)
if not args.output_hardsubs and not args.input_path.is_file():
    args.output_hardsubs = compatibility_path if (compatibility_path := Path(args.input_path.parent, "legen_burned_" + args.input_path.name)).exists() else Path(args.input_path.parent, "hardsubs_" + args.input_path.name)

if args.transcription_device == "auto":
    import torch

    matmul_backend = getattr(torch.backends.cuda, "matmul", None)
    if matmul_backend is not None and hasattr(matmul_backend, "fp32_precision"):
        matmul_backend.fp32_precision = "tf32"
    elif matmul_backend is not None and hasattr(matmul_backend, "allow_tf32"):
        matmul_backend.allow_tf32 = True

    cudnn_conv_backend = getattr(torch.backends.cudnn, "conv", None)
    if cudnn_conv_backend is not None and hasattr(cudnn_conv_backend, "fp32_precision"):
        cudnn_conv_backend.fp32_precision = "tf32"
    elif hasattr(torch.backends.cudnn, "allow_tf32"):
        torch.backends.cudnn.allow_tf32 = True
    torch_device = ("cuda" if torch.cuda.is_available() else "cpu")
else:
    torch_device = str.lower(args.transcription_device)

transcription_compute_type = args.transcription_compute_type if args.transcription_compute_type != "default" else "float16" if not torch_device == "cpu" else "float32"

args.transcription_model = "large-v3" if args.transcription_model == "large" else args.transcription_model

# ----------------------------------------------------------------------------

if args.norm:
    # normalize video using vidqa
    with time_task(message_start=f"Running {wblue}vidqa{default} and updating folder modifiation times in {gray}{args.input_path}{default}", end="\n"):
        subprocess.run(["vidqa", "-i", args.input_path, "-m", "unique", "-fd",
                        Path(Path(getframeinfo(currentframe()).filename).resolve().parent, "vidqa_data")])
        # update folder time structure
        file_utils.update_folder_times(args.input_path)

# load whisper model
with time_task(message_start=f"\nLoading {args.transcription_engine} model: {wblue}{args.transcription_model}{default} ({transcription_compute_type}) on {wblue}{torch_device}{default}", end="\n"):
    if args.transcription_engine == 'whisperx':
        import whisperx
        import whisperx_utils
    
        whisper_model = whisperx.load_model(
            whisper_arch=args.transcription_model, device=torch_device, compute_type=transcription_compute_type, asr_options={"repetition_penalty": 1, "prompt_reset_on_temperature": 0.5, "no_repeat_ngram_size": 2,})
    elif args.transcription_engine == 'whisper':
        import whisper

        import whisper_utils
        whisper_model = whisper.load_model(
            name=args.transcription_model, device=torch_device, in_memory=True)
    else:
        raise ValueError(f'Unsupported transcription engine {args.transcription_engine}. Supported values: whisperx, whisper')

with time_task(message="⌛ Processing files for"):
    path: Path
    for path in (item for item in sorted(sorted(Path(args.input_path).rglob('*'), key=lambda x: x.stat().st_mtime), key=lambda x: len(x.parts)) if item.is_file()):
        rel_path = path.relative_to(args.input_path)
        with time_task(message_start=f"\nProcessing {yellow}{rel_path.as_posix()}{default}", end="\n", message="⌚ Done in"):
            try:
                # define file type by extensions
                if path.suffix.lower() in video_extensions:
                    file_type = "video"
                elif path.suffix.lower() in audio_extensions:
                    file_type = "audio"
                else:
                    file_type = "other"

                if file_type == "video" or file_type == "audio":
                    # define paths
                    origin_media_path = path
                    dupe_filename = len(check_other_extensions(path, list(video_extensions | audio_extensions))) > 1
                    posfix_extension = path.suffix.lower().replace('.', '_') if dupe_filename else ''

                    softsub_video_dir = Path(args.output_softsubs, rel_path.parent)
                    burned_video_dir = Path(args.output_hardsubs, rel_path.parent)
                    # output video extension will be changed to .mp4
                    softsub_video_path = Path(args.output_softsubs, rel_path.stem + posfix_extension + ".mp4")
                    hardsub_video_path = Path(burned_video_dir, rel_path.stem + posfix_extension + ".mp4")
                    subtitle_translated_path = Path(
                        softsub_video_dir, rel_path.stem + posfix_extension + f"_{args.translate}.srt")
                    subtitles_path = []

                    if args.input_lang == "auto":
                        # extract audio
                        audio_short_extracted = file_utils.TempFile(
                            None, file_ext=".wav")
                        ffmpeg_utils.extract_short_wav(
                            origin_media_path, audio_short_extracted.getpath())
                        # detect language
                        print("Detecting audio language: ", end='', flush=True)
                        if args.transcription_engine == 'whisperx':
                            audio_language = whisperx_utils.detect_language(
                                whisper_model, audio_short_extracted.getpath())
                        if args.transcription_engine == 'whisper':
                            audio_language = whisper_utils.detect_language(
                                whisper_model, audio_short_extracted.getpath())
                        print(f"{gray}{audio_language}{default}")

                        audio_short_extracted.destroy()
                    else:
                        audio_language = args.input_lang
                        print(f"Forced input audio language: {gray}{audio_language}{default}")
                    # set path after get transcribed language
                    subtitle_transcribed_path = Path(
                        softsub_video_dir, rel_path.stem + posfix_extension + f"_{audio_language}.srt")
                    # create temp file for .srt
                    transcribed_srt_temp = file_utils.TempFile(
                        subtitle_transcribed_path, file_ext=".srt")
                    # skip transcription if transcribed srt for this language is existing (without overwrite neabled) or will not be used in LeGen process
                    if (file_utils.file_is_valid(subtitle_transcribed_path)) or ((args.disable_hardsubs or file_utils.file_is_valid(hardsub_video_path)) and (args.disable_srt or file_utils.file_is_valid(subtitle_transcribed_path))) and not args.overwrite:
                        print("Transcription is unnecessary. Skipping.")
                    else:
                        # extract audio
                        audio_extracted = file_utils.TempFile(None, file_ext=".wav")
                        ffmpeg_utils.extract_audio_wav(
                            origin_media_path, audio_extracted.getpath())
                        # transcribe saving subtitles to temp .srt file
                        if args.transcription_engine == 'whisperx':
                            print(f"{wblue}Transcribing{default} with {gray}WhisperX{default}")
                            whisperx_utils.transcribe_audio(
                                whisper_model, audio_extracted.getpath(), transcribed_srt_temp.getpath(), audio_language, device=torch_device, batch_size=args.transcription_batch)
                        if args.transcription_engine == 'whisper':
                            print(f"{wblue}Transcribing{default} with {gray}Whisper{default}")
                            whisper_utils.transcribe_audio(
                                model=whisper_model, audio_path=audio_extracted.getpath(), srt_path=transcribed_srt_temp.getpath(), lang=audio_language, disable_fp16=False if transcription_compute_type == "float16" or transcription_compute_type == "fp16" else True)

                        audio_extracted.destroy()
                        # if save .srt is enabled, save it to destination dir, also update path with language code
                        if not args.disable_srt:
                            transcribed_srt_temp.save()
                    transcribed_srt_source_path = transcribed_srt_temp.getvalidpath()
                    if transcribed_srt_source_path:
                        subtitles_path.append(transcribed_srt_source_path)
                    # translate transcribed subtitle using Google Translate if transcribed language is not equals to target
                    # skip translation if translation has not requested, has equal source and output language, if file is existing (without overwrite neabled) or will not be used in LeGen process
                    if args.translate == "none":
                        translated_srt_source_path = None
                        pass # translation not requested
                    elif args.translate == audio_language:
                        print("Translation is unnecessary because input and output language are the same. Skipping.")
                        translated_srt_source_path = None
                    elif (args.disable_hardsubs or file_utils.file_is_valid(hardsub_video_path)) and (args.disable_srt or (file_utils.file_is_valid(subtitle_translated_path) and file_utils.file_is_valid(subtitle_transcribed_path) and file_utils.file_is_valid(subtitle_translated_path))) and not args.overwrite:
                        print("Translation is unnecessary. Skipping.")
                        subtitles_path.insert(0, subtitle_translated_path)
                        translated_srt_source_path = subtitle_translated_path if file_utils.file_is_valid(subtitle_translated_path) else None
                    elif file_utils.file_is_valid(subtitle_translated_path):
                        print("Translated file found. Skipping translation.")
                        subtitles_path.insert(0, subtitle_translated_path)
                        translated_srt_source_path = subtitle_translated_path
                    elif transcribed_srt_temp.getvalidpath():
                        # create the temp .srt translated file
                        translated_srt_temp = file_utils.TempFile(
                            subtitle_translated_path, file_ext=".srt")

                        # translating subtitle using Google Translate or Gemini
                        print(f"{wblue}Translating{default} with {gray}{args.translate_engine.capitalize()}{default} to {gray}{args.translate}{default}")
                        subs = translate_utils.translate_srt_file(
                            transcribed_srt_temp.getvalidpath(),
                            translated_srt_temp.getpath(),
                            args.translate,
                            translate_engine=args.translate_engine,
                            gemini_api_keys=args.gemini_api_keys
                        )
                        if not args.disable_srt:
                            translated_srt_temp.save()

                        translated_srt_source_path = translated_srt_temp.getvalidpath()
                        subtitles_path.insert(0, translated_srt_source_path)
                    else:
                        translated_srt_source_path = None
                    if args.export_txt and transcribed_srt_source_path:
                        export_txt_if_requested(transcribed_srt_source_path, subtitle_transcribed_path.with_suffix(".txt"))
                    if args.export_txt and translated_srt_source_path:
                        export_txt_if_requested(translated_srt_source_path, subtitle_translated_path.with_suffix(".txt"))
                    if not args.disable_softsubs:
                        if file_utils.file_is_valid(softsub_video_path) and not args.overwrite:
                            print(f"Existing video file {gray}{softsub_video_path}{default}. Skipping subtitle insert")
                        else:
                            # create the temp .mp4 with srt in video container
                            video_softsubs_temp = file_utils.TempFile(
                                softsub_video_path, file_ext=".mp4")

                            # insert subtitle into container using ffmpeg
                            print(f"{wblue}Inserting subtitle{default} in mp4 container using {gray}FFmpeg{default}")
                            ffmpeg_utils.insert_subtitle(input_media_path=origin_media_path, subtitles_path=subtitles_path,
                                                        burn_subtitles=False, output_video_path=video_softsubs_temp.getpath(),
                                                        codec_video=args.codec_video, codec_audio=args.codec_audio)
                            video_softsubs_temp.save()
                    if not args.disable_hardsubs:
                        if file_utils.file_is_valid(hardsub_video_path) and not args.overwrite:
                            print(f"Existing video file {gray}{hardsub_video_path}{default}. Skipping subtitle burn")
                        else:
                            # create the temp .mp4 with srt in video container
                            video_hardsubs_temp = file_utils.TempFile(
                                hardsub_video_path, file_ext=".mp4")
                            # insert subtitle into container and burn using ffmpeg
                            print(f"{wblue}Inserting subtitle{default} in mp4 container and {wblue}burning{default} using {gray}FFmpeg{default}")
                            ffmpeg_utils.insert_subtitle(input_media_path=origin_media_path, subtitles_path=subtitles_path,
                                                        burn_subtitles=True, output_video_path=video_hardsubs_temp.getpath(),
                                                        codec_video=args.codec_video, codec_audio=args.codec_audio)
                            video_hardsubs_temp.save()
                else:
                    print("not a video file")
                    if args.copy_files:
                        if not args.disable_srt:
                            # copia o arquivo extra para pasta que contém também os arquivos srt
                            file_utils.copy_file_if_different(path, Path(
                                args.output_softsubs, rel_path))
                        if not args.disable_hardsubs:
                            # copia o arquivo extra para pasta que contém os videos queimados
                            file_utils.copy_file_if_different(path, Path(
                                args.output_hardsubs, rel_path))
            except Exception as e:
                file = path.as_posix()
                print(f"{red}ERROR !!!{default} {file}")
                print(f"{yellow}check legen-errors.txt for details{default}")
                # extract the relevant information from the exception object
                current_time = time.strftime("%y/%m/%d %H:%M:%S", time.localtime())

                error_message = f"[{current_time}] {file}: {type(e).__name__}: {str(e)}"
                # write the error message to a file
                with open(Path(Path(getframeinfo(currentframe()).filename).resolve().parent, "legen-errors.txt"), "a") as f:
                    f.write(error_message + "\n")
                    f.close()

    print("Deleting temp folder")
    file_utils.delete_folder(
        Path(Path(getframeinfo(currentframe()).filename).resolve().parent, "temp"))

    print(f"{green}Tasks done!{default}")
